@@ -2,9 +2,10 @@ package portmap
 
 import (
 	"fmt"
-	"github.com/containernetworking/cni/pkg/types/current"
-	"github.com/greenpau/cni-plugins/pkg/utils"
 	"net"
+
+	current "github.com/containernetworking/cni/pkg/types/040"
+	"github.com/greenpau/cni-plugins/pkg/utils"
 )
 
 // Interface represents a collection of addresses
@@ -217,30 +218,9 @@ func (p *Plugin) execAdd(conf *Config, prevResult *current.Result) error {
 	// Set bridge interface name
 	bridgeIntfName := p.interfaceChain[0]
 
-	// Add pre-routing rules
 	for _, targetInterface := range p.targetInterfaces {
 		for _, addr := range targetInterface.addrs {
-			chainName := utils.GetChainName("npr", conf.ContainerID)
-			exists, err := utils.IsChainExists(addr.Version, p.natTableName, chainName)
-			if err != nil {
-				return fmt.Errorf(
-					"failed obtaining ipv%s prerouting %s chain info: %s",
-					addr.Version, chainName, err,
-				)
-			}
-			if !exists {
-				if err := utils.CreateChain(
-					addr.Version,
-					p.natTableName,
-					chainName,
-					"none", "none", "none",
-				); err != nil {
-					return fmt.Errorf(
-						"failed creating ipv%s prerouting %s chain: %s",
-						addr.Version, chainName, err,
-					)
-				}
-			}
+
 			if len(conf.RuntimeConfig.PortMaps) == 0 {
 				continue
 			}
@@ -250,17 +230,6 @@ func (p *Plugin) execAdd(conf *Config, prevResult *current.Result) error {
 			if addr.Version == "6" && conf.ContIPv6.String() == "" {
 				continue
 			}
-			if err := utils.CreateJumpRule(
-				addr.Version,
-				p.natTableName,
-				p.preRoutingNatChainName,
-				chainName,
-			); err != nil {
-				return fmt.Errorf(
-					"failed creating jump rule from ipv%s prerouting %s chain: %s",
-					addr.Version, chainName, err,
-				)
-			}
 
 			var destAddr net.IPNet
 			if addr.Version == "4" {
@@ -269,12 +238,74 @@ func (p *Plugin) execAdd(conf *Config, prevResult *current.Result) error {
 				destAddr = conf.ContIPv6
 			}
 
+			nprChain := utils.GetChainName("npr", conf.ContainerID)
+			npoChain := utils.GetChainName("npo", conf.ContainerID)
+
+			// Add NPR chain.
+			if exists, err := utils.IsChainExists(addr.Version, p.natTableName, nprChain); !exists && err == nil {
+				if err := utils.CreateChain(
+					addr.Version,
+					p.natTableName,
+					nprChain,
+					"none", "none", "none",
+				); err != nil {
+					return fmt.Errorf(
+						"failed creating ipv%s prerouting %s chain: %s",
+						addr.Version, nprChain, err,
+					)
+				}
+			} else if err != nil {
+				return fmt.Errorf(
+					"failed obtaining ipv%s prerouting %s chain info: %s",
+					addr.Version, nprChain, err,
+				)
+			}
+
+			// Add postrouting chain
+			if exists, err := utils.IsChainExists(addr.Version, p.natTableName, npoChain); !exists && err == nil {
+				if err := utils.CreateChain(
+					addr.Version,
+					p.natTableName,
+					npoChain,
+					"none", "none", "none",
+				); err != nil {
+					return fmt.Errorf(
+						"failed creating ipv%s postrouting %s chain: %s",
+						addr.Version, npoChain, err,
+					)
+				}
+			} else if err != nil {
+				return fmt.Errorf(
+					"failed obtaining ipv%s postrouting %s chain info: %s",
+					addr.Version, npoChain, err,
+				)
+			}
+
+			if r, err := utils.GetJumpRule(addr.Version, p.natTableName, p.postRoutingNatChainName, npoChain); err == nil && r == nil {
+				if err := utils.CreateJumpRule(
+					addr.Version,
+					p.natTableName,
+					p.postRoutingNatChainName,
+					npoChain,
+				); err != nil {
+					return fmt.Errorf(
+						"failed creating jump rule to ipv%s postrouting %s chain: %s",
+						addr.Version, npoChain, err,
+					)
+				}
+			} else if err != nil {
+				return fmt.Errorf(
+					"failed check for jump rule to ipv%s postrouting %s chain: %s",
+					addr.Version, npoChain, err,
+				)
+			}
+
 			for _, pm := range conf.RuntimeConfig.PortMaps {
 				if err := utils.AddDestinationNatRules(
 					map[string]interface{}{
 						"version":          addr.Version,
 						"table":            p.natTableName,
-						"chain":            chainName,
+						"chain":            nprChain,
 						"bridge_interface": bridgeIntfName,
 						"ip_address":       destAddr,
 						"port_mapping":     pm,
@@ -282,28 +313,10 @@ func (p *Plugin) execAdd(conf *Config, prevResult *current.Result) error {
 				); err != nil {
 					return fmt.Errorf(
 						"failed creating destination NAT rules in %s chain of %s table for %v: %s",
-						chainName, p.natTableName, pm, err,
+						nprChain, p.natTableName, pm, err,
 					)
 				}
 
-				/*
-					if err := utils.AddDestinationNatRewriteRules(
-						map[string]interface{}{
-							"version":          addr.Version,
-							"table":            p.rawTableName,
-							"chain":            p.preRoutingRawChainName,
-							"bridge_interface": bridgeIntfName,
-							"ip_address":       destAddr,
-							"port_mapping":     pm,
-						},
-					); err != nil {
-						return fmt.Errorf(
-							"failed creating destination NAT rewrite rules in %s chain of %s table for %v: %s",
-							p.preRoutingRawChainName, p.rawTableName, pm, err,
-						)
-					}
-				*/
-
 				// Check whether the rule allowing traffic to leave out of
 				// bridge interface, e.g. cni-podman0, exists.
 				// If it does not exist, create it.
@@ -322,27 +335,94 @@ func (p *Plugin) execAdd(conf *Config, prevResult *current.Result) error {
 						addr.Version, p.forwardFilterChainName, p.filterTableName, pm, err,
 					)
 				}
-
 			}
 
+			// Add postrouting masquerade into the container bridge network.
+			if err := utils.AddPostRoutingDestNatRule(
+				map[string]interface{}{
+					"version":          addr.Version,
+					"table":            p.natTableName,
+					"chain":            npoChain,
+					"bridge_interface": bridgeIntfName,
+					"ip_address":       destAddr,
+				},
+			); err != nil {
+				return fmt.Errorf(
+					"failed creating postrouting rule for localhost ipv%s %s chain of %s table: %s",
+					addr.Version, p.forwardFilterChainName, p.filterTableName, err,
+				)
+			}
+
+			// Loop through local IP addresses for the next two rule
+			// creation functions. The loops are blatently stolen from
+			// https://stackoverflow.com/questions/23558425/how-do-i-get-the-local-ip-address-in-go
+			hostInterfaces, err := net.Interfaces()
+			if err != nil {
+				return fmt.Errorf("Failed to get local interfaces: %s", err)
+			}
+
+			for _, i := range hostInterfaces {
+
+				// Skip the container bridge interface
+				if i.Name == bridgeIntfName {
+					continue
+				}
+
+				hostIPAddrs, err := i.Addrs()
+				if err != nil {
+					return fmt.Errorf(
+						"Failed to get IP addresses for interface %s: %s",
+						i.Name, err,
+					)
+				}
+				for _, hostIPAddr := range hostIPAddrs {
+					var hostAddr net.IP
+					switch foo := hostIPAddr.(type) {
+					case *net.IPNet:
+						hostAddr = foo.IP
+					case *net.IPAddr:
+						hostAddr = foo.IP
+					}
+
+					// Skip IPv6 addresses when working with IPv4, and vice versa.
+					if addr.Version == "4" && hostAddr.To4() == nil {
+						continue
+					}
+					if addr.Version == "6" && hostAddr.To16() == nil {
+						continue
+					}
+
+					// Add an `ip daddr` jump rule to the NAT prerouting chain.
+					if err := utils.CreateJumpRuleWithIPDaddrMatch(
+						addr.Version,
+						p.natTableName,
+						p.preRoutingNatChainName,
+						nprChain,
+						hostAddr,
+					); err != nil {
+						return fmt.Errorf(
+							"failed creating jump rule from ipv%s prerouting %s chain: %s",
+							addr.Version, nprChain, err,
+						)
+					}
+
+					// Add an `ip daddr` jump rule to the NAT output chain.
+					if err := utils.CreateJumpRuleWithIPDaddrMatch(
+						addr.Version,
+						p.natTableName,
+						p.outputNatChainName,
+						nprChain,
+						hostAddr,
+					); err != nil {
+						return fmt.Errorf(
+							"failed creating jump rule from ipv%s output %s chain: %s",
+							addr.Version, nprChain, err,
+						)
+					}
+				}
+			}
 		}
 	}
-
-	/*
-	   RuntimeConfig: (struct { PortMaps []portmap.MappingEntry "json:\"portMappings,omitempty\"" }) {
-	    PortMaps: ([]portmap.MappingEntry) (len=1 cap=4) {
-	     (portmap.MappingEntry) {
-	      HostPort: (int) 46063,
-	      ContainerPort: (int) 80,
-	      Protocol: (string) (len=3) "tcp",
-	      HostIP: (string) ""
-	     }
-	    }
-	   },
-	   ContIPv4: (net.IPNet) 10.88.0.7/16,
-	   ContIPv6: (net.IPNet) <nil>
-	*/
-
 	return nil
 }
 
@@ -462,91 +542,143 @@ func (p *Plugin) execCheck(conf *Config, prevResult *current.Result) error {
 				v, p.forwardFilterChainName, p.filterTableName,
 			)
 		}
-
 	}
 
 	return nil
 }
 
 func (p *Plugin) execDelete(conf *Config, prevResult *current.Result) error {
+	var err error
+	var natTableExists, filterTableExists, forwardFilterChainExists, preRoutingNatChainExists, postRoutingNatChainExists, outputNatChainExists, nprExists, npoExists bool
+
 	if err := p.validateInput(conf, prevResult); err != nil {
 		return fmt.Errorf("failed validating input: %s", err)
 	}
 
-	for v := range p.targetIPVersions {
-		exists, err := utils.IsTableExist(v, p.natTableName)
-		if err != nil {
-			return fmt.Errorf("failed obtaining ipv%s nat table %s info: %s", v, p.natTableName, err)
-		}
-		if !exists {
-			continue
-		}
-		exists, err = utils.IsChainExists(v, p.natTableName, p.preRoutingNatChainName)
-		if err != nil {
-			return fmt.Errorf(
-				"failed obtaining ipv%s prerouting chain %s info: %s",
-				v, p.preRoutingNatChainName, err,
-			)
-		}
-		if exists {
-			for _, targetInterface := range p.targetInterfaces {
-				for _, addr := range targetInterface.addrs {
-					if v != addr.Version {
-						continue
-					}
-					chainName := utils.GetChainName("npr", conf.ContainerID)
-					if err := utils.DeleteJumpRule(addr.Version, p.natTableName, p.preRoutingNatChainName, chainName); err != nil {
-						return err
-					}
-				}
-			}
-		}
-	}
-
+	nprChain := utils.GetChainName("npr", conf.ContainerID)
+	npoChain := utils.GetChainName("npo", conf.ContainerID)
 	bridgeIntfName := p.interfaceChain[0]
 
-	for _, targetInterface := range p.targetInterfaces {
-		for _, addr := range targetInterface.addrs {
-			chainName := utils.GetChainName("npr", conf.ContainerID)
-			exists, err := utils.IsChainExists(addr.Version, p.natTableName, chainName)
-			if err != nil {
-				continue
+	for v := range p.targetIPVersions {
+
+		if natTableExists, err = utils.IsTableExist(v, p.natTableName); natTableExists && err == nil {
+			if preRoutingNatChainExists, err = utils.IsChainExists(v, p.natTableName, p.preRoutingNatChainName); err != nil {
+				return fmt.Errorf(
+					"error checking ipv%s prerouting chain %s info: %s",
+					v, p.preRoutingNatChainName, err,
+				)
 			}
-			if exists {
-				if err := utils.DeleteChain(addr.Version, p.natTableName, chainName); err != nil {
-					return err
+
+			if postRoutingNatChainExists, err = utils.IsChainExists(v, p.natTableName, p.postRoutingNatChainName); err != nil {
+				return fmt.Errorf(
+					"error checking ipv%s postrouting chain %s info: %s",
+					v, p.postRoutingNatChainName, err,
+				)
+			}
+
+			if outputNatChainExists, err = utils.IsChainExists(v, p.natTableName, p.outputNatChainName); err != nil {
+				return fmt.Errorf(
+					"error checking ipv%s output chain %s info: %s",
+					v, p.outputNatChainName, err,
+				)
+			}
+		} else if err != nil {
+			return fmt.Errorf(
+				"error checking ipv%s nat table %s info: %s",
+				v, p.natTableName, err,
+			)
+		}
+
+		if filterTableExists, err = utils.IsTableExist(v, p.filterTableName); filterTableExists && err == nil {
+			if forwardFilterChainExists, err = utils.IsChainExists(v, p.filterTableName, p.forwardFilterChainName); err != nil {
+				return fmt.Errorf(
+					"error checking ipv%s forward filter chain %s info: %s",
+					v, p.forwardFilterChainName, err,
+				)
+			}
+		} else if err != nil {
+			return fmt.Errorf(
+				"error checking ipv%s filter table %s info: %s",
+				v, p.filterTableName, err,
+			)
+		}
+
+		for _, targetInterface := range p.targetInterfaces {
+			for _, addr := range targetInterface.addrs {
+				if v != addr.Version {
+					continue
 				}
-			}
 
-			var destAddr net.IPNet
-			if addr.Version == "4" {
-				destAddr = conf.ContIPv4
-			} else {
-				destAddr = conf.ContIPv6
-			}
-
-			for _, pm := range conf.RuntimeConfig.PortMaps {
-
-				// Remove filter forwarding rules
-				if err := utils.RemoveFilterForwardMappedPortRules(
-					map[string]interface{}{
-						"version":          addr.Version,
-						"table":            p.filterTableName,
-						"chain":            p.forwardFilterChainName,
-						"bridge_interface": bridgeIntfName,
-						"ip_address":       destAddr,
-						"port_mapping":     pm,
-					},
-				); err != nil {
+				if nprExists, err = utils.IsChainExists(addr.Version, p.natTableName, nprChain); err != nil {
 					return fmt.Errorf(
-						"failed removing filter forward mapped port rules in ipv%s %s chain of %s table for %v: %s",
-						addr.Version, p.forwardFilterChainName, p.filterTableName, pm, err,
+						"error checking ipv%s prerouting container chain %s info: %s",
+						v, nprChain, err,
 					)
 				}
-			}
 
+				if npoExists, err = utils.IsChainExists(addr.Version, p.natTableName, npoChain); err != nil {
+					return fmt.Errorf(
+						"error checking ipv%s postrouting container chain %s info: %s",
+						v, npoChain, err,
+					)
+				}
+
+				if natTableExists {
+					if nprExists {
+						if preRoutingNatChainExists {
+							if err := utils.DeleteJumpRule(addr.Version, p.natTableName, p.preRoutingNatChainName, nprChain); err != nil {
+								return err
+							}
+						}
+						if outputNatChainExists {
+							if err := utils.DeleteJumpRule(addr.Version, p.natTableName, p.outputNatChainName, nprChain); err != nil {
+								return err
+							}
+						}
+						if err := utils.DeleteChain(addr.Version, p.natTableName, nprChain); err != nil {
+							return err
+						}
+					}
+					if npoExists {
+						if postRoutingNatChainExists {
+							if err := utils.DeleteJumpRule(addr.Version, p.natTableName, p.postRoutingNatChainName, npoChain); err != nil {
+								return err
+							}
+						}
+						if err := utils.DeleteChain(addr.Version, p.natTableName, npoChain); err != nil {
+							return err
+						}
+					}
+				}
+
+				var destAddr net.IPNet
+				if addr.Version == "4" {
+					destAddr = conf.ContIPv4
+				} else {
+					destAddr = conf.ContIPv6
+				}
+
+				if filterTableExists && forwardFilterChainExists {
+					for _, pm := range conf.RuntimeConfig.PortMaps {
+						if err := utils.RemoveFilterForwardMappedPortRules(
+							map[string]interface{}{
+								"version":          addr.Version,
+								"table":            p.filterTableName,
+								"chain":            p.forwardFilterChainName,
+								"bridge_interface": bridgeIntfName,
+								"ip_address":       destAddr,
+								"port_mapping":     pm,
+							},
+						); err != nil {
+							return fmt.Errorf(
+								"failed removing filter forward mapped port rules in ipv%s %s chain of %s table for %v: %s",
+								addr.Version, p.forwardFilterChainName, p.filterTableName, pm, err,
+							)
+						}
+					}
+				}
+			}
 		}
 	}
-
 	return nil
 }
